# packages needed 
import os
import re
import networkx as nx

import matplotlib.pyplot as plt
from networkx.drawing.nx_agraph import graphviz_layout

# plotting colored data on map
import numpy as np
import pandas as pd # if load coordinate data using pd
from mpl_toolkits.basemap import Basemap # if plot map


def getResultFilesInfo(PATH_TO_RESULT_FILES):
    '''
    Obtain descriptive information of the communities from file names of NetStruct Hierarhcy's output files.

    Parameters:
        PATH_TO_RESULT_FILES (str): Path to the NetStruct Hierarhcy's output directory.

    Returns:
        fileInfo (list): A list of descriptive information of the communities in the hierarchy. Each item is a list containing info on one community, including its level, entry, parent level, parent entry, parent line, and threshold value (separated into two parts before and after the decimal point).
    '''
    
    # get files from the directory
    files = [f for f in os.listdir(PATH_TO_RESULT_FILES) if re.match(r'Le_[0-9]+_En_[0-9]+_PaLe_[0-9]+_PaEn_[0-9]+_PaLi_[0-9]+_TH_.*_C.*', f)]
    file0 = [f for f in os.listdir(PATH_TO_RESULT_FILES) if re.match(r'Le_0_En_0_TH_*',f)][0]
    
    # retrieve community info from the file name
    fileInfo = list()
    file0_info = re.findall(r'\d+', file0)
    fileInfo.append( (file0_info[0:2] + [None, None, None] + file0_info[2:]) )

    for f in files:
        fileInfo.append(re.findall(r'\d+', f))

    fileInfo = sorted(fileInfo, key=lambda x: int(x[0]))

    return fileInfo


def extractCommInfo(fileInfo, PATH_TO_RESULT_FILES):
    '''
    Extract detailed community information from the NetStruct Hierarhcy's output files.

    Parameters:
        fileInfo (list): A list of descriptive information of the communities in the hierarchy. Can be obtained using getResultFilesInfo().
        PATH_TO_RESULT_FILES (str): Path to the NetStruct Hierarhcy's output directory.

    Returns:
        belonging_map (dict): A map of each node (community) in the hierarchy to the individuals it contained. Key is the node label, and value is the list of individuals.
        parent_map (dict): A map of each node (community) in the hierarchy to its parent node (community in one level up). Key is the node label, and value is the parent node label.
        child_map (dict): A map of each node (community) in the hierarchy to its child nodes (communities in one level down). The key is the node labels, and the values are parent node labels.
    '''
                    
    belonging_map = dict() # map of node to individuals belonging to the node
    parent_map = dict() # map of node to parent of the node
    child_map = dict() # map of node to the list of children of the node

    for fileIdx in range(len(fileInfo)):
        # load the output file for each level of the hierarchy
        info = fileInfo[fileIdx]
        le, en = info[0], info[1]
        if le=='0':
            nodeFile = "Le_0_En_0_TH_0.0000000_C.txt"
        else:
            nodeFile = "Le_{}_En_{}_PaLe_{}_PaEn_{}_PaLi_{}_TH_{}.{}_C.txt".format(le,en,info[2],info[3],info[4],info[5],info[6])

        f = open(PATH_TO_RESULT_FILES+nodeFile, 'r')
        lines = f.readlines()
        f.close()

        # get individuals in each sub-community
        s_raw = [l.split() for l in lines]
        s_all = [list([int(i) for i in subs]) for subs in s_raw] 

        # examine all communitites split from the node
        l_comms = list()
        for comm_idx in range(len(s_all)):
            comm = s_all[comm_idx]
            clusterID = '-'.join([le,en,str(comm_idx)]) # node label
            
            belonging_map[clusterID] = comm
            
            if fileIdx == 0:
                parent_map[clusterID] = 'ROOT'
                if 'ROOT' not in child_map:
                    child_map['ROOT'] = [clusterID]
                else:
                    child_map['ROOT'].append(clusterID)
            else:
                par = '-'.join(info[2:5]) # parent node
                parent_map[clusterID] = par
                if par not in child_map:
                    child_map[par] = [clusterID]
                else:
                    child_map[par].append(clusterID)

    return belonging_map,parent_map,child_map


def buildTree(belonging_map,parent_map,child_map,root='ROOT'):
    '''
    Construct the hierarchy from NetStruct Hierarhcy's output as a tree (using graph).

    Parameters:
        belonging_map (dict): The map of each node (community) in the hierarchy to the individuals it contained.
        parent_map (dict): The map of each node (community) in the hierarchy to its parent node.
        child_map (dict): The map of each node (community) in the hierarchy to its child nodes.
        root (str): Label of the root node (optional; default: 'ROOT')

    Returns:
        G (Graph): A tree representation of the hierarchy. Each node is a community, and each edge represents a subset relationship between two communities at adjacent levels (the community at lower level is a subset of that at the higher level).
        color_int_map (dict): A map of each node (community) in the hierarchy to the color interval [left, right] using the gradient color scheme. 
    '''
    
    # build the tree and color map
    
    internal = [root]
    color_int_map = dict()
    color_int_map[root] = [0,1]

    G = nx.DiGraph()
    G.add_node(root)
    
    subn = len(child_map[root])
    ca = color_int_map[root][0]
    cb = color_int_map[root][1]

    for i,c in enumerate(child_map[root]):
        G.add_node(c)
        G.add_edge(root, c)
        color_int_map[c] = [ca+(cb-ca)*(i)/subn, ca+(cb-ca)*(i+1)/subn]

    # recursively build the tree
    below_cur = child_map[root].copy()
    while len(below_cur) != 0:
        nextCluster = below_cur.pop()
        
        if nextCluster in child_map: # continue if not leaf node
            children = child_map[nextCluster]
            G.add_node(nextCluster)
            
            subn = len(children)
            ca = color_int_map[nextCluster][0]
            cb = color_int_map[nextCluster][1]
            
            for i,c in enumerate(children):
                G.add_node(c)
                G.add_edge(nextCluster, c)
                color_int_map[c] = [ca+(cb-ca)*(i)/subn, ca+(cb-ca)*(i+1)/subn]
                
            below_cur.extend(children)

    return G, color_int_map


def plotTree(G, color_int_map, PATH_TO_SAVE_PLOTS="", save_plot=False, plot_name = "", **kwargs):
    '''
    Plot the hierarchy from NetStruct Hierarhcy's output as a tree.

    Parameters:
        G (Graph): A tree representation of the hierarchy.
        color_int_map (dict): A map of each node (community) in the hierarchy to the color interval.
        PATH_TO_SAVE_PLOTS (str): Path to the saving directory (optional; default: "")
        save_plot (bool): Whether to save the plot (optional; default: False)
        plot_name (str): Name of the plot (optional; default: "")
        scale_nodes (bool): Whether to scale each node by the community size (optional; default: False)
        belonging_map (dict): The map of each node (community) in the hierarchy to the individuals it contained (optinal; default: None; only required if scale_nodes)
        label_nodes (bool): Whether to label each node (optional; default: True)
        color_nodes (bool): Whether to color each node by the gradient color scheme (optional; default: True)
        
      --Other optional parameters in plotting:
        figsize: default (5,5)
        dpi: default 100
        rect_axes: default [0,0,1,1]
        axis_aspect: default 1
        title: default ""
        node_size: default 300
        node_alpha: default 0.9
        label_size: default 9
        
    Returns:
        fig (Figure): A figure of the hierarchy. 
    '''
    
    # get the input parameters
    figsize = kwargs.pop('figsize', (5,5))
    dpi = kwargs.pop('dpi', 100)
    rect_axes = kwargs.pop('rect_axes', [0,0,1,1])
    axis_aspect = kwargs.pop('axis_aspect', 1)
    title = kwargs.pop('title', "")
    node_size = kwargs.pop('node_size', 300)
    node_alpha = kwargs.pop('node_alpha', 0.9)
    scale_nodes = kwargs.pop('scale_nodes', False)
    belonging_map = kwargs.pop('belonging_map', None)
    label_nodes = kwargs.pop('label_nodes', True)
    label_size = kwargs.pop('label_size', 8)
    color_nodes = kwargs.pop('color_nodes', True)  
    
    fig=plt.figure(figsize=figsize, dpi=dpi)
    plt.box(False); plt.axis('off')
    plt.tick_params(top='off', bottom='off', left='off', right='off', labelleft='off', labelbottom='off')
    plt.title(title)
    pos =graphviz_layout(G, prog='dot') # positions of the nodes
    ax=plt.axes(rect_axes); ax.set_aspect(axis_aspect)
    # plot the tree
    nx.draw_networkx_edges(G, pos, ax=ax)
    nx.draw(G, pos, ax=ax, with_labels=False, arrows=True, node_size=1)

    # plot colored nodes
    if color_nodes:
        cmap = plt.cm.rainbow

    if scale_nodes:
        tot_size = len(G.nodes)
    else:
        size = node_size
    for node in G.nodes():
        x, y= pos[node]
        col = (0,0,0,0.2)  #default: grey
        if color_nodes:
            col = cmap((color_int_map[node][0]+color_int_map[node][1])/2)
        if scale_nodes:
            if node=="ROOT":
                size = 1
            else:
                size = len(belonging_map[node])/tot_size*node_size
        ax.scatter(x, y, c=[col], s=size, alpha = node_alpha, zorder=2)
        if label_nodes:
            ax.text(x, y, node, c = (0,0,0,0.9), fontsize=label_size)

    if save_plot:
        plt.savefig(PATH_TO_SAVE_PLOTS+plot_name,dpi=dpi,bbox_inches = 'tight') 
    
    return fig


def plotMap(G, belonging_map, color_int_map, latitudes, longitudes, label_inds = False, ind_labels = None, PATH_TO_SAVE_PLOTS="", save_plot=False, plot_name = "", **kwargs):
    
    '''
    Plot the individuals on the map, colored using the hierarchy output by NetStruct.

    Parameters:
        G (Graph): A tree representation of the hierarchy.
        belonging_map (dict): The map of each node (community) in the hierarchy to the individuals it contained.
        color_int_map (dict): A map of each node (community) in the hierarchy to the color interval.
        latitudes (list): The list of latitude coordinate of each individual.
        longitudes (list): The list of longitude coordinate of each individual.
        PATH_TO_SAVE_PLOTS (str): Path to the saving directory (optional; default: "")
        save_plot (bool): Whether to save the plot (optional; default: False)
        plot_name (str): Name of the plot (optional; default: "")
        label_inds (bool): Whether to label each individual (optional; default: False)
        ind_labels (list): The list of individual labels.
        
      --Other optional parameters in plotting:
        figsize: default (5,5)
        dpi: default 100
        marker_size: default 40
        marker_alpha: default 0.9
        label_size: default 6
        map_bdry: default {"llcrnrlat":-60, "urcrnrlat":75,"llcrnrlon":-175, "urcrnrlon":175}
        
    Returns:
        fig (Figure): A figure of the map. 
    '''
    
    # get the input parameters
    figsize = kwargs.pop('figsize', (9,6))
    dpi = kwargs.pop('dpi', 100)
    
    marker_size = kwargs.pop('marker_size', 40)
    marker_alpha = kwargs.pop('marker_alpha', 0.9)
    label_size = kwargs.pop('label_size', 6)  
    map_bdry = kwargs.pop('map_bdry', {"llcrnrlat":-60, "urcrnrlat":75,"llcrnrlon":-175, "urcrnrlon":175})

    latitudes = np.array(latitudes)
    longitudes = np.array(longitudes)
    
    # background map
    fig = plt.figure(figsize=figsize,dpi=dpi)
    m = Basemap(projection='merc', resolution='c',
                llcrnrlat=map_bdry["llcrnrlat"], urcrnrlat=map_bdry["urcrnrlat"], 
                llcrnrlon=map_bdry["llcrnrlon"], urcrnrlon=map_bdry["urcrnrlon"])
    m.drawcoastlines()
    m.fillcontinents(color='lightgray',lake_color='white')
    
    lats = m.drawparallels(np.linspace(-90, 90, 7),labels=[True,True,False,False],color=(0,0,0,0.1))
    lons = m.drawmeridians(np.linspace(-180, 180, 7),labels=[False,False,False,True],color=(0,0,0,0.1))

    # plot individuals in the map
    cmap = plt.cm.rainbow
    interested_clusters = set()
    for node in G.nodes():

        all_comms_at_level = belonging_map[node]
        interested_clusters |= set(all_comms_at_level)
        col = cmap((color_int_map[node][0]+color_int_map[node][1])/2)   
        m.scatter(longitudes[all_comms_at_level], latitudes[all_comms_at_level], latlon=True, 
                  c = [col], s=[marker_size], alpha=marker_alpha, zorder=2)

    if label_inds:
        for idx in range(len(ind_labels)):
            x, y = m(longitudes[idx], latitudes[idx])
            plt.annotate(ind_labels[idx], xy=(x,y), xytext=(1,1), xycoords='data',textcoords ='offset points',
                        fontsize=label_size, zorder=2)

    if save_plot:
        plt.savefig(PATH_TO_SAVE_PLOTS+plot_name,dpi=dpi,bbox_inches = 'tight') 
        
    return fig